#include "defs.h"
#include "block.h"
#include "array_m.h"
#include "vector.h"
#include "hilbert.h"
#include "spins.h"
#ifdef CONFIG_PTHREAD
#include <pthread.h>
#endif
#ifdef CONFIG_MPI
#include <mpi.h>
#endif

struct vblock {     /* arguments for working threads */
  int blk;          /* block (thread) nr */
  mcplx *v0;        /* vector (used by clrvec,mltply,addav) */
  double r;         /* factor (used by mltply,addav) */
  mcplx *v1;        /* vector (used by addav) */
};

/* void-function for pthreads */
void copy_av_block( void *ptr ){
  int   blk=((struct vblock *)ptr)->blk;
  mcplx *v0=((struct vblock *)ptr)->v0+b_ofs[blk];
  mcplx *v1=((struct vblock *)ptr)->v1+b_ofs[blk];
  double  r=((struct vblock *)ptr)->r, n2t=0;
  ULong i;
  for(i=0l;i<b_ofs[blk+1]-b_ofs[blk];i++) {
    v1[i]  = M_d2f(M_f2d(v0[i])*r);
    // n2t   += Norm(M_f2d(v1[i]));
  }
  ((struct vblock *)ptr)->r = n2t; /* scalar product (v1,v1) */
  return;
}

double copy_av(tvec *v1, double r, tvec *v0){  /* v1=r*v0, [return norm(v1)] v2.27 */
  struct vblock buf[B_NUM];
  int blk; double n2=0;
  for(blk=0;blk<pt_n;blk++) {
    buf[blk].blk=blk;
    buf[blk].r  =r;
    buf[blk].v0 =v0;
    buf[blk].v1 =v1;
    parallel_start( blk, copy_av_block, &buf[blk]);
  }
  for (blk=0;blk<pt_n;blk++) {
    parallel_wait(blk); /* synchronize */
    n2 += buf[blk].r;
  }
#if 0
#ifdef CONFIG_MPI
  { double add=n2, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    n2=sum; }
#endif
#endif
  return(n2);
}

/* void-function for pthreads */
void add_av_block( void *ptr ){
  int   blk=((struct vblock *)ptr)->blk;
  mcplx *v0=((struct vblock *)ptr)->v0+b_ofs[blk];
  mcplx *v1=((struct vblock *)ptr)->v1+b_ofs[blk];
  double  r=((struct vblock *)ptr)->r, n2t=0;
  ULong i;
  for(i=0l;i<b_ofs[blk+1]-b_ofs[blk];i++) {
    v1[i] += M_d2f(M_f2d(v0[i])*r);
    n2t   +=  Norm(M_f2d(v1[i]));
  }
  ((struct vblock *)ptr)->r = n2t; /* scalar product (v1,v1) */
  return;
}

double add_av(tvec *v1, double r, tvec *v0){  /* v1+=r*v0, return norm(v1) */
  struct vblock buf[B_NUM];
  int blk; double n2=0;
  for(blk=0;blk<pt_n;blk++) {
    buf[blk].blk=blk;
    buf[blk].r  =r;
    buf[blk].v0 =v0;
    buf[blk].v1 =v1;
    parallel_start( blk, add_av_block, &buf[blk]);
  }
  for (blk=0;blk<pt_n;blk++) {
    parallel_wait(blk); /* synchronize */
    n2 += buf[blk].r;
  }
#ifdef CONFIG_MPI
  { double add=n2, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    n2=sum; }
#endif
  return(n2);
}

/* void-function for pthreads */
void mltply_av_block( void *ptr ){
  int   blk=((struct vblock *)ptr)->blk;
  mcplx *v0=((struct vblock *)ptr)->v0+b_ofs[blk];
  double  r=((struct vblock *)ptr)->r;
  ULong i;
  for(i=0l;i<b_ofs[blk+1]-b_ofs[blk];i++) 
    v0[i] = M_d2f(M_f2d(v0[i])*r);
  return;
}

void mltply_av(double r, tvec *v0){  /* v0*=r */
  struct vblock buf[B_NUM];
  int blk;
  for(blk=0;blk<pt_n;blk++) {
    buf[blk].blk=blk;
    buf[blk].r  =r;
    buf[blk].v0 =v0;
    parallel_start( blk, mltply_av_block, &buf[blk]);
  }
  for (blk=0;blk<pt_n;blk++) parallel_wait(blk); /* synchronize */
  return;
}

/* void-function for pthreads */
void clrvec_block( void *ptr ){
  int  blk=((struct vblock *)ptr)->blk;
  mcplx *v=((struct vblock *)ptr)->v0+b_ofs[blk];
  ULong i;
  for (i=0l;i<b_ofs[blk+1]-b_ofs[blk];i++) v[i]=M_d2f(0);
  return;
}

void clrvec(tvec *v){ int blk;
  /* its important to inititialize memory locally in the threads for NUMA */
  struct vblock buf[B_NUM];
  if (vvv && !mpi_i) printf("\n clrvec");
  if (vvv) MPI_Barrier(MPI_COMM_WORLD);
  // JS-2012-07 add error check (MPI.sum_ERR?), 2012-10 case mpi_n > n1
  if ((!v) && b_ofs[pt_n]) F0("!v"); 
  for (blk=0;blk<pt_n;blk++) {
#if 0
    if (vvv&4 && (mpi_i*pt_n+blk<4 || mpi_i*pt_n+blk>=mpi_n*pt_n-4))
       printf(" v_%d_%d[%lu]=0", mpi_i, blk, b_ofs[blk+1]-b_ofs[blk]);
#endif
    buf[blk].blk=blk;
    buf[blk].v0 =v;
    parallel_start( blk, clrvec_block, &buf[blk]);
  }
  for (blk=0;blk<pt_n;blk++) parallel_wait(blk); /* synchronize */
  if (vvv) MPI_Barrier(MPI_COMM_WORLD);
  if (vvv && !mpi_i) printf(" ok");
  return;
}

double normvec(tvec *v){ /* non-thread-parallel, ToDo: norm_v() */
  double r=0.0; ULong j;
  for (j=0l;j<b_ofs[pt_n];j++)
    r += Norm(M_f2d(v[j]));
#ifdef CONFIG_MPI
  { double add=r, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    r=sum; }
#endif
  if (r<=1e-20){ if(vvv&4 && !mpi_i) printf(" norm_vec<1e-20"); } else
  if (r!=1.0){ r=1.0/sqrt(r); 
    if (vvv&1 && !mpi_i) printf(" normvec 1-r*(1/r)=%.0e",1-r*(1/r));
    mltply_av(r,v); // v=r*v;
    for (r=j=0l;j<b_ofs[pt_n];j++)
      r += Norm(M_f2d(v[j]));
#ifdef CONFIG_MPI
  { double add=r, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    r=sum; }
#endif
    if (vvv&1 && !mpi_i) printf(" 1-norm=%.0e",1-r);
  }
  return r;
}

/* if blk==-1 save all blocks, idx>=0 index for name="d%d.vec" */
/* called only by thread blk==0 
 * MPI: writing into one file by (open+append+close+nexttask) does not work
 *   test 2012-05-15 mkautz.nfs, so we use MPI_IO (bad data, extreme slow)
 */
int savevec(tvec *v, const char *nam, int idx, int blk){
#if CONFIG_MPI
  // MPI_Offset my_offset; // , my_current_offset;   // offset = integer?
  // MPI_File fh;
  MPI_Status status;
#endif
  int i=0, nerr=0; char s1[128], *s2;
  if (blk>0) {
    printf("\nWarning savevec: code called from working thread, skip");
    return 0;
  }
  if (idx>=0) i=sprintf(s2=s1,nam,idx); else s2=(char *)nam;
  if (i<0) E0("fname to long");
  if ((vvv&5)==1 && !mpi_i) printf(" wv");      // short form
  if ((vvv&4)==4 && !mpi_i) printf(" w.%s",s2); // long form
#if CONFIG_MPI  && 0 // 2012-06-06 disabled fails on q.NFS
  // 2012-05-15 (120qu.nfs3.ADIOI_Set_lock failed+exit, needs noac)
  //  File locking failed in ADIOI_Set_lock(fd 6,cmd F_SETLKW/7,type F_WRLCK/1,whence 0) with return value FFFFFFFF and errno 25.
  //  If the file system is NFS, you need to use NFS version 3, ensure that the lockd daemon is running on all the machines, and mount the directory with the 'noac' option (no attribute caching).
  // ToDo: transfer in 2MB chunks to task0 (io-task)
  MPI_File_open(MPI_COMM_WORLD, s2, MPI_MODE_CREATE
                                  | MPI_MODE_WRONLY, MPI_INFO_NULL,&fh);
  my_offset= mpi_ofs[mpi_i]*Csize;
  MPI_File_seek(fh, my_offset, MPI_SEEK_SET);
  // MPI_File_get_position(fh, &my_current_offset);
  if (((int)b_ofs[pt_n])*Csize!=b_ofs[pt_n]*1.*Csize) F0("int_overflow");
  MPI_File_write(fh, (void *)v, b_ofs[pt_n]*Csize, MPI_CHAR, &status);
  // MPI_Get_count(&status, MPI_CHAR, &count); // ToDo: error checking
  MPI_File_close(&fh);
  { int etmp=0;
    MPI_Allreduce(&nerr,&etmp, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
    nerr=etmp;
  }
#else // 0
  // ToDo: send to task0, task0 recvs from all tasks and writes (2MB chunks)
  // 2012-06-07 works for NFS without noac
  if (!mpi_i) {
    FILE *f=NULL; ULong n2;
    if (!(f=fopen(s2,"wb"))) { E1("fopen(%s)",s2);perror("");nerr=1; }
    // old parallel: v==0 happens for tasks where node_n1==0 (i.A. n1 < mpi_n)
    if ((!nerr) && b_ofs[pt_n] && v) {
      n2=fwrite((void *)v,Csize,(size_t)b_ofs[pt_n],f); /* full node */
      if (n2!=b_ofs[pt_n]){E2("write=%lu err=%d",n2,ferror(f));nerr=2;}
    }
#if CONFIG_MPI    // 2012-06-06 works also if locking doesnt work (NFS2,...)
    if (mpi_n>1) { // task0 recv from all other tasks
      int i1, i3, n3; ULong i2; char *buf; 
      buf=(char *)malloc(1<<24); // 16MB
      if (!buf) { E0("malloc16MB"); nerr=-2; } // ToDo: reduce on err, send bsize?
      for (i1=1; i1<mpi_n; i1++) {
        int b2= (buf)?0:-1; // 0 = ok, -1 = error abort
        // printf("DBG.t0: send token %d to i1=%d\n",b2,i1);
        MPI_Send(&b2,1,MPI_INT,i1,1,MPI_COMM_WORLD); // send token to i1
        if (buf)
        for (i2=mpi_ofs[i1]*Csize; i2<mpi_ofs[i1+1]*Csize; i2+=1<<24) {
          if (i2+(1<<24)<=mpi_ofs[i1+1]*Csize) i3=1<<24;
          else                                 i3=mpi_ofs[i1+1]*Csize-i2;
          MPI_Recv(buf,i3,MPI_BYTE,/*src*/i1,/*tag*/0,MPI_COMM_WORLD,&status);
          // MPI_Get_count(&status, MPI_BYTE, &i3 /* out.count */);
          n3=fwrite(buf,1,i3,f); /* full node */
          if (n3!=i3){ E2("write=%d err=%d",i3,ferror(f));nerr=2; }
        } // loop over transfer chunks
      } // loop over tasks
      if (buf) free(buf); // libc may still not free for later use
    } // mpi_n>1
#endif
    if (f) fclose(f);
  } // mpi_i==0
#if CONFIG_MPI    // 2012-06-06 works also if locking doesnt work (NFS2,...)
  else { // mpi_i!=0   send to task0
    ULong i2; int i3, b2;
    //  n1=80 = 0,16,32,48,64 + 80(rest0) (16 Byte chunks)
    //  n1=77 = 0,16,32,48    + 64(rest13)
    // 2012-11-14 JS problems with mpich, without token ...
    //    ... all nodes send data and n0.mpich may buffer and crash on OOM
    MPI_Recv(&b2,1,MPI_INT,0,1,MPI_COMM_WORLD,&status); // blocking
    // printf("DBG.t%d: got token %d from 0\n",mpi_i,b2);
    if (!b2) // skip if error token was send
    for (i2=0; i2<b_ofs[pt_n]*Csize; i2+=1<<24) {
      if (i2+(1<<24)<=b_ofs[pt_n]*Csize) i3=1<<24;
      else                               i3=b_ofs[pt_n]*Csize-i2;
      MPI_Send(((char *)v)+i2,i3,MPI_BYTE,/*dst*/0,/*tag*/0,MPI_COMM_WORLD);
    }
  }
#endif
#endif // MPI?
  return(nerr);
}

int loadvec(tvec *v,const char *nam,int idx,int blk){
#if CONFIG_MPI
  MPI_Status status;
#endif
  FILE *f=NULL;
  ULong n2;
  int i=0, nerr=0;
  char s1[128], *s2;
  if (blk>0) {
    printf("\nWarning savevec: code called from working thread, skip");
    return 0;
  }
  if (idx>=0) i=sprintf(s2=s1,nam,idx); else s2=(char *)nam;
  if (i<0 && !mpi_i) E0("fname to long");
  if ((vvv&5)==1 && !mpi_i) printf(" rv");      // short form
  if ((vvv&4)==4 && !mpi_i) printf(" r.%s",s2); // long form
  /* ToDo: do not return in a mpi-loop!, 2011-04-01 +nerr */
  /* multible open+append+close from nodes to nfs fail 2012-05-15 mkautz */
  /*   caused by scfabric errors? t=34533459 ??? nolock,rsize=1024,wsize=1024 */
  /* better send all data to thread0, which does the file IO (extra code + buffer)
    or use MPI_IO (ToDo 2012-05) */
#if CONFIG_MPI && 0 // 2012-05-15, same lock-problem as save
//  MPI_Offset my_offset;
//  MPI_File fh;
  MPI_File_open(MPI_COMM_WORLD, s2, MPI_MODE_RDONLY, MPI_INFO_NULL,&fh);
  my_offset= mpi_ofs[mpi_i]*Csize;
  MPI_File_seek(fh, my_offset, MPI_SEEK_SET);
  // MPI_File_get_position(fh, &my_current_offset);
  if (((int)b_ofs[pt_n])*Csize!=b_ofs[pt_n]*1.*Csize) F0("int_overflow");
  MPI_File_read(fh, (void *)v, b_ofs[pt_n]*Csize, MPI_CHAR, &status);
  // MPI_Get_count(&status, MPI_CHAR, &count); // ToDo: error checking
  MPI_File_close(&fh);
  { int etmp=0;
    MPI_Allreduce(&nerr,&etmp, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
    nerr=etmp;
  }
#else
  if (mpi_i==0) {
    if (!(f=fopen(s2,"rb"))) {E1("fopen %s",s2);perror("");nerr=1;} 
    /* Csize = sizeof(mcplx) */
    // if ((!nerr) && fseek(f,mpi_ofs[mpi_i]*Csize,SEEK_SET)){E0("fseek");nerr=4;}
    if ((!nerr) /* && b_ofs[pt_n]*/ && v) { // including read 0 members??
      n2 = fread((void *)v,Csize,(size_t)b_ofs[pt_n],f);
      if (n2!=b_ofs[pt_n]) {
        int n3=0;
        if (ferror(f)) {
          E2("fread1=%ld ferror=%d",(long)n2,ferror(f)); nerr=2; }
        if (!nerr)
          n3 = fread((void *)(v+n2),Csize,(size_t)b_ofs[pt_n]-n2,f);
        if (n2+n3!=b_ofs[pt_n] || ferror(f)) {
          E2("fread2=%ld ferror=%d",(long)n3,ferror(f)); nerr=2; }
      }
    } 
#if CONFIG_MPI    // 2012-06-06 works also if locking doesnt work (NFS2,...)
    if (mpi_n>1) { // recv from all other threads
      int i1; ULong i2, i3=0; char *buf;
      buf=(char *)malloc(1<<24); // 16MB
      if (!buf) E0("malloc");
      if (buf) for (i1=1; i1<mpi_n; i1++)
       for (i2=mpi_ofs[i1]*Csize; i2<mpi_ofs[i1+1]*Csize; i2+=1<<24) {
        if (i2+(1<<24)<=mpi_ofs[i1+1]*Csize) i3=1<<24;
        else                                 i3=mpi_ofs[i1+1]*Csize-i2;
        n2=fread(buf,1,i3,f); /* full node */
        if (n2!=i3){ E2("fread=%ld ferror=%d",(long)n2,ferror(f));nerr=2; }
        MPI_Send(buf,i3,MPI_BYTE,/*dst*/i1,/*tag*/0,MPI_COMM_WORLD);
      } // loop^2
      if (buf) free(buf);
    } // mpi_n>1
#endif
    if (f) fclose(f);
  } // mpi_i==0
#if CONFIG_MPI    // 2012-06-06 works also if locking doesnt work (NFS2,...)
  else { // mpi_i!=0   send to task0
    ULong i2, i3;
    for (i2=0; i2<b_ofs[pt_n]*Csize; i2+=1<<24) {
      if (i2+(1<<24)<=b_ofs[pt_n]*Csize) i3=1<<24;
      else                               i3=(b_ofs[pt_n]*Csize)-i2;
      MPI_Recv(((char *)v)+i2, i3, MPI_BYTE,
               /*src*/0, /*tag*/0, MPI_COMM_WORLD, &status);
    }
  }
#endif
#endif
  return(nerr);
}

int loadvec2(tvec *v,const char *nam,int idx){ /* load v from textfile */
  FILE *f; double r; ULong j; tbase i; int k,l,ret; char buf[255];
  if (mpi_n>1) F0("missing MPI code");
  int ii=0; char s1[128],*s2;
  if(idx>=0) ii=sprintf(s2=s1,nam,idx); else s2=(char *)nam;
  if(ii<0) E0("file name to long");
  if(!v){ printf(" failed, v==NULL"); return 3;}
  if(!(f=fopen(s2,"rt"))){E1("loadvec_fopen(%s)",s2);return 1;}
  clrvec(v);ret=1;
  while(!feof(f)){
      if(!fgets(buf,255,f))break; /* zeile lesen: 0.25 |ud.3u> /n */
      if(sscanf(buf,"%lf",&r)!=1)continue;
      for(k=0;buf[k];k++) if(buf[k]=='|')break;
      i=0l;j=0l;l=2*(nn-1) CM(/2,+0,+0);
      if(buf[k]=='|'){
         for(k++;buf[k]!=0 && l>=0;k++,CM(l--,l-=2,l-=2))	/* Apr99 */
         if(buf[k]=='.'){                   } else
         if(buf[k]=='u'){i|=CM(0l,1l,1l)<<l;} else
         if(buf[k]=='d'){i|=CM(1l,2l,2l)<<l;} else
         if(buf[k]=='3'){i|=         3l <<l;} else
         E0("unknown symbol (not .ud3)"); }
      else E1("no config |xxxx> found in %s",nam);   
      j=scfg2idx(i); /* 0 .. node_n1-1 */
      if(vvv)printf("\n read: %6.3f |%s> => %ld",r,ccfg(i),j);
      v[j]=M_d2f(r);ret=0;}
  fclose(f);
  return ret;
}
   
#ifndef RAND_MAX
#define RAND_MAX 0x7fff
#endif
extern int startvec;
/* --- last change: 7.12.98 JS bug! --- 
 *  sum v[] during setting for faster normalization???
 *  ini could be choosen randomly to ensure different numeric errors
 */
void randvec(tvec *v, int ini){
    ULong i;

    ini += (startvec&~7);
    if (vvv && !mpi_i) printf(" randvec%d-%d", startvec&7, ini);
    if (vvv) MPI_Barrier(MPI_COMM_WORLD);
    clrvec(v); /* memory first touch for ccNUMA */
    if (startvec==1){ /* set one component to 1, rest to 0 */
      if (  (ULong)ini>=mpi_ofs[mpi_i  ]
         && (ULong)ini< mpi_ofs[mpi_i+1]  )
        v[  (ULong)ini -mpi_ofs[mpi_i  ] ]=M_d2f(1);
      if ((ULong)ini>=mpi_ofs[mpi_n]) v[0]=M_d2f(1); /* very small n1 */
    } else
    if (startvec==2){ loadvec( v,FileGS,ini,-1);   } else
    if (startvec==3){ loadvec2(v,FileSV,ini);      } else
    if (startvec==5){ /* ToDo: set all v[i]=1 ? */
      E0("startvec=5 removed in v2.35, ToDo: storeh may return i_min");
    } else
    if (startvec>4){ printf(" no startvec"); } else
    { /* srand(rnd);rnd++; -- bad idea, resulting in a bug until v2.17 */ 
      /* ToDo: spread in a way that norm is exactly 1.0, use l^2+m^2=n^2 */
      /* ToDo1104: different results for different mpi_n, better 
       *   using mypseudorandom(index+const)
       *   or randomarray[index%len]^rarray[(index/len)%len]^const
       *   or const + some random bigger values ??? */
      srand(ini);
      for(i=0l;i<b_ofs[pt_n];i++)
        v[i]=M_d2f((1023 & rand())/1024.); /* Jul00 */
      if (!mpi_i) v[0]=M_d2f(0.5); /* be sure v is never zero */
    }
    if (vvv) MPI_Barrier(MPI_COMM_WORLD);
    if (vvv && !mpi_i) printf(" randvec=ok normvec");
    normvec(v); return;
}
    
/* scalar product, need correct order(?): <a+ib|c+id>!=<c+id|a+ib> */
cplx sprod(tvec *a, tvec *b){ /* non-thread-parallel, tvec=mcplx since v2.36 */
  cplx r=0.0; ULong i;
  if (a && b)  /* not for a2 2x2 2013-01 */
  for (i=0l;i<b_ofs[pt_n];i++) 
    r += Conj(M_f2d(a[i])) * M_f2d(b[i]);
#ifdef CONFIG_MPI
  { double add[2]={Re(r),Im(r)}, sum[2]={0,0};
    MPI_Allreduce(&add, &sum, (VecType&4)/4+1,
                  MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    r=C(sum[0],sum[1]); }
#endif
  return(r);
}
    
int  iortho(tvec *v0, tvec *v1){ /* 0=ok,1=parallel */
    cplx r1; double r2; ULong i;
    r1=sprod(v0,v1);
    if (Norm(r1)<1.0e-10) return(0);
    r2=0.0;
    for (i=0l;i<b_ofs[pt_n];i++){
      /* Conj(r1) for correct degeneracy for complex vectors */
      v0[i]-=M_d2f(Conj(r1)*M_f2d(v1[i])); /* round error ??? */
      r2+=Norm(M_f2d(v0[i]));
    }
#ifdef CONFIG_MPI
  { double add=r2, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    r2=sum; }
#endif
    /* 0.995 was successfull for degeneracy of kagome-JJ N=54 u+d=51+3 q_0
     * 0.99999 results in to high degeneracy (909 instead of 72) Jul04
     * */
    if (Norm(r1)>0.995) {printf(" iortho: v0=v1");return(1);}
    r2=1.0/sqrt(r2);
//    for (i=0l;i<b_ofs[pt_n];i++)
//      v0[i]=M_d2f(M_f2d(v0[i])*r2);  /* call mltly_av ??? */
    mltply_av(r2,v0);
    return(0);
}

